'''
..
  $Id$

  Remoting python tools of the clazzes.org project
  http://www.clazzes.org

  Created: 04.02.2018

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

from clazzes.remoting import cmd
from clazzes.remoting import serializer
from clazzes.remoting import compact
from clazzes.remoting import aux
from clazzes.tools.lang import extension

import logging

import threading
import socket

from urllib.parse import urlparse

log = logging.getLogger(__name__)

bean_type_lock = threading.Lock()

bean_types = {}

def defineMethod(name):
    def callMethod(self,*args):
        return self.client.invoke(self.subSystem,name,args)

    return callMethod

def getBeanClass(interfaceClass):

    with bean_type_lock:

        BeanClass = bean_types.get(interfaceClass)

        if BeanClass is None:

            classDict = {}

            for name in interfaceClass.__abstractmethods__:
                classDict[name] = defineMethod(name)

            def stubCtor(self,client,subSystem):
                self.client = client
                self.subSystem = subSystem

            classDict["__init__"] = stubCtor

            BeanClass = type(interfaceClass.__name__+"_stub",(interfaceClass,),classDict)

            bean_types[interfaceClass] = BeanClass

    return BeanClass

def makeBean(interfaceClass,client,subSystem):

    BeanClass = getBeanClass(interfaceClass)

    return BeanClass(client,subSystem)

class Client:
    '''
      Handle a client connection and handle remote invocations.
    '''

    def __init__(self,srlzr:serializer.Serializer):
        '''
          Set up a connection handler based on a serializer.
        '''
        self.uid = None
        self.srlzr = srlzr

    @staticmethod
    def connect(addr,serialIDs,timeout=30):
        '''
        Connect a client to the given address.
        :param addr: URL to connect to, default value is ``localhost``, where the default
                 port is ``1234``, if no port is specified in the URL. An urllib
                 ParseResult object may also be passed.
        '''
        if isinstance(addr,str):
            addr = urlparse(addr)

        sock = None

        try:
            if addr.scheme == "tcp":

                host = addr.hostname
                port = addr.port if addr.port else 1234

                addrinfos = socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM)
                for addrinfo in addrinfos:
                    family, socktype, proto, _canonname, host_port = addrinfo
                    try:
                        sock = socket.socket(family,socktype,proto)
                        sock.settimeout(timeout)
                        sock.connect(host_port)
                        break

                    except (OSError, IOError):
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug('Client.connect(%s:%d) failed'%(host_port[0],host_port[1]),exc_info=True)
                        sock.close()
                        sock = None

                if sock is None:
                    raise IOError("Cannot connect to TCP socket [%s:%d]."%(host,port))

            elif addr.scheme == "unix":
                sock = socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
                sock.connect(addr.path)

            else:
                raise ValueError("Invalid scheme [%s] for remoting URL."%addr.scheme)

            ret = Client(compact.CompactSerializer(sock.makefile(mode="rwb"),serialIDs))

            ret.waitForServer()

            sock = None
            return ret

        finally:
            if sock is not None:
                sock.close()

    def waitForServer(self):

        c = self.srlzr.readObject("__cmd")

        if not isinstance(c,cmd.ExchangeClientUIDCmd):
            raise aux.RemoteInvocationFailureException("Received invalid command [%s] at connection startup."%type(cmd))

        self.uid = c.clientUID

    def invoke(self,subSystem,methodName,args):

        if self.uid is None:
            raise AssertionError("invoke() called before waitForServer().")

        invocationCmd = cmd.InvocationCmd()
        invocationCmd.subSystem = subSystem
        invocationCmd.arg = aux.RemoteInvocation()

        invocationCmd.arg.methodName = methodName
        invocationCmd.arg.arguments = args

        self.srlzr.writeObject("__invocationCmd",invocationCmd)

        self.srlzr.flush()

        c = self.srlzr.readObject("__cmd")

        return c.receive_client()

    def close(self):
        self.srlzr.close()

    def __enter__(self):
        return self

    def __exit__(self,*args): # @UnusedVariable
        self.close()
        return False

@extension(cmd.InvocationResultCmd)
def receive_client(self):

    return self.result

@extension(cmd.InvocationExceptionCmd)
def receive_client(self):  # @DuplicatedSignature

    raise self.exception
