'''
..
  $Id$
  
  Remoting python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 04.02.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import sys
import logging
from clazzes.remoting import limits
from clazzes.remoting import stream
from clazzes.remoting import serializer
import numbers

from collections.abc import ByteString,Mapping,Set,Sequence

log = logging.getLogger(__name__)

class UnexpectedTagError(Exception):
    pass


class CompactSerializer(serializer.Serializer):
    '''
    A serializer operating  on a binary file handle an reading/writing
    the clazzes.org compact tag-value stream format.
    '''

    '''
     Stream format.
     <pre>
     +------ 0 -------+
     | NULL_STREAM_ID |
     +----------------+
     </pre>
    '''
    NULL_STREAM_ID               = 0x00

    '''
      Stream format.
      <pre>
      +-------- 0 --------+-- 1 --+
      | BOOLEAN_STREAM_ID | {0|1} |
      +-------------------+-------+
      </pre>
     '''
    BOOLEAN_STREAM_ID            = 0x01

    '''
      Stream format.
      <pre>
      +-------- 0 --------+- 1 --+
      | BOOLEAN_STREAM_ID | byte |
      +-------------------+------+
      </pre>
     '''
    BYTE_STREAM_ID               = 0x02

    '''
      Stream format.
      <pre>
      +------- 0 -------+--- 1-2 ----+
      | SHORT_STREAM_ID | short(MSB) |
      +-----------------+------------+
      </pre>
     '''
    SHORT_STREAM_ID              = 0x03

    '''
      Stream format.
      <pre>
      +------- 0 -------+-- 1-4 ---+
      | SHORT_STREAM_ID | int(MSB) |
      +-----------------+----------+
      </pre>
     '''
    INT_STREAM_ID                = 0x04

    '''
      Stream format.
      <pre>
      +------- 0 -------+--- 1-8 ---+
      | SHORT_STREAM_ID | long(MSB) |
      +-----------------+-----------+
      </pre>
     '''
    LONG_STREAM_ID               = 0x05

    '''
      Stream format.
      <pre>
      +------- 0 -------+------- 1-4 --------+
      | SHORT_STREAM_ID | float(ieee754/MSB) |
      +-----------------+--------------------+
      </pre>
     '''
    FLOAT_STREAM_ID              = 0x06

    '''
      Stream format.
      <pre>
      +------- 0 -------+-------- 1-8 --------+
      | SHORT_STREAM_ID | double(ieee754/MSB) |
      +-----------------+---------------------+
      </pre>
     '''
    DOUBLE_STREAM_ID             = 0x07

    '''
      Stream format.
      <pre>
      +------- 0 -------+--- 1-2 ---+
      | SHORT_STREAM_ID | char(MSB) |
      +-----------------+-----------+
      </pre>
     '''
    CHAR_STREAM_ID               = 0x08

    '''
      Stream format.
      <pre>
      +----------- 0 -----------+-------- 1-8 ---------+- 9-12 -+----- 13-(13+n) ------+
      | BOOLEAN_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | b[0],b[1],...,b[n-1] |
      +-------------------------+----------------------+--------+----------------------+
      </pre>
     '''
    BOOLEAN_ARRAY_STREAM_ID      = 0x09

    '''
      Stream format.
      <pre>
      +---------- 0 ---------+-------- 1-8 ---------+- 9-12 -+----- 13-(13+n) ------+
      | BYTE_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | b[0],b[1],...,b[n-1] |
      +----------------------+----------------------+--------+----------------------+
      </pre>
     '''
    BYTE_ARRAY_STREAM_ID         = 0x0a

    '''
      Stream format.
      <pre>
      +---------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+2*n) -----+
      | SHORT_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | s[0],s[1],...,s[n-1] |
      +-----------------------+----------------------+--------+----------------------+
      s[j]...MSB short
      </pre>
     '''
    SHORT_ARRAY_STREAM_ID        = 0x0b

    '''
      Stream format.
      <pre>
      +--------- 0 ---------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+4n) -----+
      | INT_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | i[0],i[1],...,i[n-1] |
      +---------------------+----------------------+--------+----------------------+
      i[j]...MSB int
      </pre>
     '''
    INT_ARRAY_STREAM_ID          = 0x0c;

    '''
      Stream format.
      <pre>
      +--------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+8*n) -----+
      | LONG_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | l[0],l[1],...,l[n-1] |
      +----------------------+----------------------+--------+----------------------+
      l[j]...MSB long
      </pre>
     '''
    LONG_ARRAY_STREAM_ID         = 0x0d

    '''
      Stream format.
      <pre>
      +---------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+4*n) -----+
      | FLOAT_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | f[0],f[1],...,f[n-1] |
      +-----------------------+----------------------+--------+----------------------+
      f[j]...ieee754/MSB float
      </pre>
     '''
    FLOAT_ARRAY_STREAM_ID        = 0x0e

    '''
      Stream format.
      <pre>
      +---------- 0 -----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+8*n) -----+
      | DOUBLE_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | d[0],d[1],...,d[n-1] |
      +------------------------+----------------------+--------+----------------------+
      d[j]...ieee754/MSB double
      </pre>
     '''
    DOUBLE_ARRAY_STREAM_ID       = 0x0f

    '''
      Stream format.
      <pre>
      +--------- 0 ----------+-------- 1-8 ---------+- 9-12 -+---- 13-(13+2*n) -----+
      | CHAR_ARRAY_STREAM_ID | object ID (MSB long) | n(MSB) | c[0],c[1],...,c[n-1] |
      +----------------------+----------------------+--------+----------------------+
      c[j]...MSB char
      </pre>
     '''
    CHAR_ARRAY_STREAM_ID         = 0x10

    '''
      Stream format.
      <pre>
      +-------- 0 -------+-------- 1-8 ---------+----- 9-10 ------+----- 11-(11+n) ------+
      | STRING_STREAM_ID | object ID (MSB long) | n(unsigned MSB) | n modified UTF bytes |
      +------------------+----------------------+-----------------+----------------------+
      </pre>
     
      @see DataOutput#writeUTF(String)
     '''
    STRING_STREAM_ID             = 0x11

    '''
      Object with full class header.
      <pre>
      +------- 0 --------+--------- 1-(n+3) ---------+------ (n+4)-(n+11) --------+
      | OBJECT_STREAM_ID | writeUTF(class.getName()) | serialVerionUID (MSB long) |
      +------------------+---------------------------+----------------------------+
     
      +----- (n+12)-(n+19) ------+--- (n+20)-(n+27) ----+
      | class info ID (MSB long) | object ID (MSB long) |
      +--------------------------+----------------------+
     
      +---------+---------+----------+-----------+
      | field 0 | field 2 |  ...     | field m-1 |
      +---------+---------+----------+-----------+
     
      n...length of modified UTF bytes in class name.
      m...number of fields written.
     
      fields are written using their individual data representation. The order of fields
      is according to the lexicographic ordering of the field names.
      </pre>
     
      @see DataOutput#writeUTF(String)
     '''
    OBJECT_STREAM_ID             = 0x12

    '''
      Object with referenced class header.
      <pre>
      +--------- 0 -----------+---------- 1-8 -----------+-------- 9-16 --------+
      | OBECT_CIREF_STREAM_ID | class info ID (MSB long) | object ID (MSB long) |
      +-----------------------+--------------------------+----------------------+
     
      +---------+---------+----------+-----------+
      | field 0 | field 2 |  ...     | field m-1 |
      +---------+---------+----------+-----------+
     
      m...number of fields written.
     
      fields are written using their individual data representation with an individual tag.
      The class info ID must have been previously serialized using
      OBJECT_STREAM_ID or OBJECT_ARRAY_STREAM_ID.
      </pre>
     '''
    OBJECT_CIREF_STREAM_ID       = 0x13

    '''
      Reference to a previously serialized Object.
      <pre>
      +--------- 0 ----------+-------- 1-8 ---------+
      | OBJECT_REF_STREAM_ID | object ID (MSB long) |
      +----------------------+----------------------+
     
      The object ID must have been previously serialized using OBJECT_STREAM_ID,
      OBJECT_CIREF_STREAM_ID, OBJECT_ARRAY_STREAM_ID, OBJECT_ARRAY_CIREF_STREAM_ID,
      LIST_STREAM_ID, SET_STREAM_ID or MAP_STREAM_IP
      </pre>
     '''
    OBJECT_REF_STREAM_ID         = 0x14

    '''
      Object array with full component class header.
      <pre>
      +---------- 0 -----------+--------- 1-(n+3) ---------+------ (n+4)-(n+11) --------+
      | OBJECT_ARRAY_STREAM_ID | writeUTF(class.getName()) | serialVerionUID (MSB long) |
      +------------------------+---------------------------+----------------------------+
     
      +----- (n+12)-(n+19) ------+--- (n+20)-(n+27) ----+- (n+28)-(n+31) -+
      | class info ID (MSB long) | object ID (MSB long) |    m(MSB int)   |
      +--------------------------+----------------------+-----------------+
     
      +------+------+-------+--------+
      | o[0] | o[1] |  ...  | o[m-1] |
      +------+------+-------+--------+
     
      n...length of modified UTF bytes in component class name.
      m...number of array members written.
     
      members are written using their individual data representation with an individual tag.
      The header class info is the array's component class.
      </pre>
     
      @see DataOutput#writeUTF(String)
      @see Array#newInstance(Class, int)
     '''
    OBJECT_ARRAY_STREAM_ID       = 0x15

    '''
      Object array with referenced component class header.
      <pre>
      +------------ 0 --------------+---------- 1-8 -----------+-------- 9-16 --------+
      | OBECT_ARRAY_CIREF_STREAM_ID | class info ID (MSB long) | object ID (MSB long) |
      +-----------------------------+--------------------------+----------------------+
     
      +-- 17-20 ---+------+------+-------+--------+
      | m(MSB int) | o[0] | o[1] |  ...  | o[m-1] |
      +------------+------+------+-------+--------+
     
      m...number of array members written.
     
      members are written using their individual data representation with an individual tag.
      The class info ID of the array's component class must have been previously serialized
      using OBJECT_STREAM_ID or OBJECT_ARRAY_STREAM_ID.
      </pre>
     '''
    OBJECT_ARRAY_CIREF_STREAM_ID = 0x16

    '''
      Stream format. Flushes the input object cache.
      <pre>
      +---------- 0 -----------+
      | OBJECT_FLUSH_STREAM_ID |
      +------------------------+
      </pre>
     '''
    OBJECT_FLUSH_STREAM_ID       = 0x17

    '''
      Optimized serialization of a list.
      <pre>
      +------ 0 -------+-------- 1-8 ---------+
      | LIST_STREAM_ID | object ID (MSB long) |
      +----------------+----------------------+
     
      +--- 9-12 ---+------+------+-------+--------+
      | m(MSB int) | o[0] | o[1] |  ...  | o[m-1] |
      +------------+------+------+-------+--------+
     
      members are written using their individual data representation with an individual tag.
      </pre>
     '''
    LIST_STREAM_ID               = 0x18

    '''
      Optimized serialization of a set.
      <pre>
      +------ 0 ------+-------- 1-8 ---------+
      | SET_STREAM_ID | object ID (MSB long) |
      +---------------+----------------------+
     
      +--- 9-12 ---+------+------+-------+--------+
      | m(MSB int) | o[0] | o[1] |  ...  | o[m-1] |
      +------------+------+------+-------+--------+
     
      members are written using their individual data representation with an individual tag.
      </pre>
     '''
    SET_STREAM_ID                = 0x19

    '''
      Optimized serialization of a map.
      <pre>
      +------ 0 ------+-------- 1-8 ---------+
      | MAP_STREAM_ID | object ID (MSB long) |
      +---------------+----------------------+
     
      +--- 9-12 ---+------+------+------+------+-------+--------+--------+
      | m(MSB int) | k[0] | v[0] | k[1] | v[1] |  ...  | k[m-1] | v[m-1] |
      +------------+------+------+------+------+-------+--------+--------+
     
      key/value pairs are written using their individual data representation with an individual tag.
      </pre>
     '''
    MAP_STREAM_ID                = 0x20

    '''
      Stream format for strings, which contain more than 65535 modified UTF-8 bytes.
      Please note, that both the length of the string in chars and the length of the
      serialized UTF bytes are written in order to allow an optimized deserialization
      implementation which allocates minimal system resources.
      <pre>
      +-------- 0 -------+-------- 1-8 ---------+---- 9-12 -----+---- 13-16 ----+----- 17-(17+n) ------+
      | STRING_STREAM_ID | object ID (MSB long) | string length | n(signed MSB) | n modified UTF bytes |
      +------------------+----------------------+---------------+---------------+----------------------+
      </pre>
     '''
    LONG_STRING_STREAM_ID        = 0x21

    
    SIMPLE_TAG_FUNCTIONS = {
        BOOLEAN_STREAM_ID:   "readByte",
        BYTE_STREAM_ID:   "readByte",
        SHORT_STREAM_ID:  "readShort",
        INT_STREAM_ID:    "readInt",
        LONG_STREAM_ID:   "readLong",
        FLOAT_STREAM_ID:  "readFloat",
        DOUBLE_STREAM_ID: "readDouble",
        CHAR_STREAM_ID:   "readChar",
        BOOLEAN_ARRAY_STREAM_ID:   "readByteArray",
        BYTE_ARRAY_STREAM_ID:   "readByteArray",
        SHORT_ARRAY_STREAM_ID:  "readShortArray",
        INT_ARRAY_STREAM_ID:    "readIntArray",
        LONG_ARRAY_STREAM_ID:   "readLongArray",
        FLOAT_ARRAY_STREAM_ID:  "readFloatArray",
        DOUBLE_ARRAY_STREAM_ID: "readDoubleArray",
        CHAR_ARRAY_STREAM_ID:   "readCharArray"
    }
    
    NUMBER_CLASS_INFOS = {
        "java.lang.Number": (numbers.Number, serializer.ClassInfo("java.lang.Number",-8742448824652078965)),
        "java.lang.Boolean": (bool,serializer.ClassInfo("java.lang.Boolean",-3665804199014368530)),
        "java.lang.Byte": (int,serializer.ClassInfo("java.lang.Byte",-7183698231559129828)),
        "java.lang.Short": (int,serializer.ClassInfo("java.lang.Short",7515723908773894738)),
        "java.lang.Integer": (int,serializer.ClassInfo("java.lang.Integer",1360826667806852920)),
        "java.lang.Long": (int,serializer.ClassInfo("java.lang.Long",4290774380558885855)),
        "java.lang.Float": (float,serializer.ClassInfo("java.lang.Float",-2671257302660747028)),
        "java.lang.Double": (float,serializer.ClassInfo("java.lang.Double",-9172774392245257468)),
        "java.lang.Character": (str,serializer.ClassInfo("java.lang.Character",3786198910865385080))
    }
    
    def __init__(self,fh,serialIDs):
        '''
        :param fh: The binary file-like object to read from and/or write to.
        :param serialIDs: A dictionary which maps serial IDs to serializable classes.
        '''
        self.stream = stream.BinaryDataStream(fh)
        self.serialIDs = serialIDs
        self.nextCID = hash("a")
        self.nextOID = hash("b")
        self.writeContext = []
        self.writeClassInfos = {}
        self.writeCache = {}
        self.readContext = []
        self.readClassInfos = {}
        self.readCache = {}
    
    def _make_oid(self):
        ret = self.nextOID
        self.nextOID = (self.nextOID + 1) & sys.maxsize
        return ret
    
    def _make_cid(self):
        ret = self.nextCID
        self.nextCID = (self.nextCID + 1) & sys.maxsize
        return ret

    def _expect_tag(self,expected,allowObject=False):
        
        tag = self.stream.readByte()
        while tag != expected:
            if tag == CompactSerializer.OBJECT_FLUSH_STREAM_ID:
                self.readCache.clear()
                tag = self.stream.readByte()
                
            elif tag == CompactSerializer.NULL_STREAM_ID:
                return None
            elif allowObject and tag == CompactSerializer.OBJECT_REF_STREAM_ID:
                
                return self.readCache[self.stream.readLong()]
            else:
                raise UnexpectedTagError("Unexpected tag [%02X] where tag [%02X] was expected."%(tag,expected))
        
        return True
    
    def _write_cached(self,o):
        
        if o is None:
            self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            return True
        
        # retrieve pair of OID and object maybe previously entered by _cache_write()
        oidoop = self.writeCache.get(id(o))
        
        if oidoop is None:
            return False
        else:
            self.stream.writeByte(CompactSerializer.OBJECT_REF_STREAM_ID)
            self.stream.writeLong(oidoop[0])
            return True

    def _cache_write(self,o):
        oid = self._make_oid()
        # enter the object ID and a reference to the object in order to
        # keep the object alive until flush() is called.
        self.writeCache[id(o)] = (oid,o)
        self.stream.writeLong(oid)

    def _get_context(self,context,field):
        
        ret = ""
        
        for ctxt in context:
            if isinstance(ctxt,numbers.Number):
                ret += "[%d]" % ctxt
            elif len(ret) > 0:
                ret += "." + ctxt
            else:
                ret += ctxt
        
        if isinstance(field,numbers.Number):
            ret += "[%d]" % field
        elif len(ret) > 0:
            ret += "." + field
        else:
            ret += field
                
        return ret

    def getReadContext(self,field):
        return self._get_context(self.readContext,field)
    
    def getWriteContext(self,field):
        return self._get_context(self.writeContext,field)

    def readBoolean(self,field):
        try:
            if self._expect_tag(CompactSerializer.BOOLEAN_STREAM_ID) is None:
                return None
            else:
                return self.stream.readByte() != 0
         
        except Exception as e:
            raise serializer.DeserializationError("Error reading byte field [%s]"%self.getReadContext(field)) from e

    def readByte(self,field):
        try:
            if self._expect_tag(CompactSerializer.BYTE_STREAM_ID) is None:
                return None
            else:
                return self.stream.readByte()
         
        except Exception as e:
            raise serializer.DeserializationError("Error reading byte field [%s]"%self.getReadContext(field)) from e

    def readChar(self,field):
        try:
            if self._expect_tag(CompactSerializer.CHAR_STREAM_ID) is None:
                return None
            else:
                return self.stream.readChar()
            
        except Exception as e:
            raise serializer.DeserializationError("Error reading char field [%s]"%self.getReadContext(field)) from e

    def readShort(self,field):
        try:
            if self._expect_tag(CompactSerializer.SHORT_STREAM_ID) is None:
                return None
            else:
                return self.stream.readShort()

        except Exception as e:
            raise serializer.DeserializationError("Error reading short field [%s]"%self.getReadContext(field)) from e

    def readInt(self,field):
        try:
            if self._expect_tag(CompactSerializer.INT_STREAM_ID) is None:
                return None
            else:
                return self.stream.readInt()

        except Exception as e:
            raise serializer.DeserializationError("Error reading int field [%s]"%self.getReadContext(field)) from e
        
    def readLong(self,field):
        try:
            if self._expect_tag(CompactSerializer.LONG_STREAM_ID) is None:
                return None
            else:
                return self.stream.readLong()

        except Exception as e:
            raise serializer.DeserializationError("Error reading long field [%s]"%self.getReadContext(field)) from e
        
    def readFloat(self,field):
        try:
            if self._expect_tag(CompactSerializer.FLOAT_STREAM_ID) is None:
                return None
            else:
                return self.stream.readFloat()
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading float field [%s]"%self.getReadContext(field)) from e
        
    def readDouble(self,field):
        try:
            if self._expect_tag(CompactSerializer.DOUBLE_STREAM_ID) is None:
                return None
            else:
                return self.stream.readDouble()
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading float field [%s]"%self.getReadContext(field)) from e
    
    def readBooleanArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.BOOLEAN_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ba = self.stream.readByteArray()
                ret = [x != 0 for x in ba]
                self.readCache[oid] = ret
                return ret
            else:
                return o
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading byte array field [%s]"%self.getReadContext(field)) from e

    def readByteArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.BYTE_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readByteArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading byte array field [%s]"%self.getReadContext(field)) from e

    def readShortArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.SHORT_ARRAY_STREAM_ID,True)
           
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readShortArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading short array field [%s]"%self.getReadContext(field)) from e

    def readIntArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.INT_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readIntArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading int array field [%s]"%self.getReadContext(field)) from e

    def readLongArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.LONG_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readLongArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o
                    
        except Exception as e:
            raise serializer.DeserializationError("Error reading long array field [%s]"%self.getReadContext(field)) from e

    def readCharArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.CHAR_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readCharArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading char array field [%s]"%self.getReadContext(field)) from e

    def readFloatArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.FLOAT_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readFloatArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o

        except Exception as e:
            raise serializer.DeserializationError("Error reading float array field [%s]"%self.getReadContext(field)) from e

    def readDoubleArray(self,field):
        try:
            o = self._expect_tag(CompactSerializer.DOUBLE_ARRAY_STREAM_ID,True)
            
            if o == True:
                oid = self.stream.readLong()
                ret = self.stream.readDoubleArray()
                self.readCache[oid] = ret
                return ret
            else:
                return o
        
        except Exception as e:
            raise serializer.DeserializationError("Error reading double array field [%s]"%self.getReadContext(field)) from e

    def readObject(self, field):
        
        try:
            tag = self.stream.readByte()
        
            while tag == CompactSerializer.OBJECT_FLUSH_STREAM_ID:
                self.readCache.clear()
                tag = self.stream.readByte()
        
            stf = CompactSerializer.SIMPLE_TAG_FUNCTIONS.get(tag)
        
            if stf is not None:
                
                readFunction = getattr(self.stream,stf)
                
                if stf.endswith("Array"):
                    oid = self.stream.readLong()
                    ret = readFunction()
                    
                    if tag == CompactSerializer.BOOLEAN_ARRAY_STREAM_ID:
                        r = [x != 0 for x in ret]
                        self.readCache[oid] = r
                    else:
                        self.readCache[oid] = ret
                        return ret
                    
                else:
                    
                    if tag == CompactSerializer.BOOLEAN_STREAM_ID:
                        return readFunction() != 0
                    else:
                        return readFunction()
        
            if tag == CompactSerializer.NULL_STREAM_ID:
                
                return None
            
            elif tag == CompactSerializer.STRING_STREAM_ID:
                
                oid = self.stream.readLong()
                ret = self.stream.readUTF()
                self.readCache[oid] = ret
                return ret
                
            elif tag == CompactSerializer.LONG_STRING_STREAM_ID:
                
                oid = self.stream.readLong()
                ret = self.stream.readUTFLong()
                self.readCache[oid] = ret
                return ret
            
            elif tag == CompactSerializer.OBJECT_STREAM_ID:
                
                serialID = self.stream.readUTF()
                serialVersionUID = self.stream.readLong()
                cid = self.stream.readLong()
                oid = self.stream.readLong()
                
                cls = self.serialIDs[serialID]

                if not issubclass(cls,serializer.Serializable):
                    raise ValueError("Class of type [%s] with serial ID [%s] is not a subclass of Serializable"%(cls,serialID))

                if serialVersionUID != cls.getSerialVersionUID() and not serialVersionUID in cls.getOldSerialVersionUIDs():
                    raise ValueError("Class with serial ID [%s] has unknown serialVersioUID [%d]"%(serialID,serialVersionUID))

                self.readClassInfos[cid] = serializer.ClassInfo(serialID,serialVersionUID)
                
                staticReader = getattr(cls,"readInstance",None)
                
                if staticReader is None:
                    ret = cls()
                
                    self.readCache[oid] = ret
                
                    self.readContext.append(field)
                
                    try:
                        ret.readFields(self,serialVersionUID)
                    finally:
                        self.readContext.pop()
                else:
                    self.readContext.append(field)
                
                    try:
                        ret = staticReader(self,serialVersionUID)
                        self.readCache[oid] = ret
                
                    finally:
                        self.readContext.pop()
                
                return ret

            elif tag == CompactSerializer.OBJECT_CIREF_STREAM_ID:
                
                cid = self.stream.readLong()
                oid = self.stream.readLong()
                
                ci = self.readClassInfos[cid]
                
                cls = self.serialIDs[ci.serialID]


                staticReader = getattr(cls,"readInstance",None)
                
                if staticReader is None:
                    ret = cls()
                
                    self.readCache[oid] = ret
                
                    self.readContext.append(field)
                
                    try:
                        ret.readFields(self,ci.serialVersionUID)
                    finally:
                        self.readContext.pop()
                else:
                    self.readContext.append(field)
                
                    try:
                        ret = staticReader(self,ci.serialVersionUID)
                        self.readCache[oid] = ret
                
                    finally:
                        self.readContext.pop()
                    
                return ret

            elif tag == CompactSerializer.OBJECT_REF_STREAM_ID:
                
                oid = self.stream.readLong()
                ret = self.readCache[oid]
                return ret

            elif tag == CompactSerializer.OBJECT_ARRAY_STREAM_ID:
                
                serialID = self.stream.readUTF()
                serialVersionUID = self.stream.readLong()
                cid = self.stream.readLong()
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of object array in stream exceeds %d"%limits.INT_MAX)

                nci = CompactSerializer.NUMBER_CLASS_INFOS.get(serialID)
                
                if nci is None:

                    cls = self.serialIDs[serialID]
    
                    if not issubclass(cls,serializer.Serializable):
                        raise ValueError("Class of type [%s] with serial ID [%s] is not a subclass of Serializable"%(cls,serialID))
    
                    if serialVersionUID != cls.getSerialVersionUID() and not serialVersionUID in cls.getOldSerialVersionUIDs():
                        raise ValueError("Class with serial ID [%s] has unknown serialVersioUID [%d]"%(serialID,serialVersionUID))

                    self.readClassInfos[cid] = serializer.ClassInfo(serialID,serialVersionUID)
                
                else:
                    cls = nci[0]
                    
                    if serialVersionUID != nci[1].serialVersionUID:
                        raise ValueError("Number class with serial ID [%s] has unknown serialVersioUID [%d]"%(serialID,serialVersionUID))
                    
                    self.readClassInfos[cid] = nci[1]
                
                ret = []
                
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    
                    for i in range(n):
                        member = self.readObject(i)
                        if not isinstance(member,cls):
                            raise ValueError("Member of type [%s] is not a subclass of [%s]"%(type(member),cls))
                        ret.append(member)
                
                finally:
                    self.readContext.pop()
                    
                return ret

            elif tag == CompactSerializer.OBJECT_ARRAY_CIREF_STREAM_ID:
                
                cid = self.stream.readLong()
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of object array in stream exceeds %d"%limits.INT_MAX)

                ci = self.readClassInfos[cid]
                
                nci = CompactSerializer.NUMBER_CLASS_INFOS.get(serialID)
                
                if nci is None:
                    cls = self.serialIDs[ci.serialID]
                else:
                    cls = nci[0]
                
                ret = []
                
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    
                    for i in range(n):
                        member = self.readObject(i)
                        if not isinstance(member,cls):
                            raise ValueError("Member of type [%s] is not a subclass of [%s]"%(type(member),cls))
                        ret.append(member)
                
                finally:
                    self.readContext.pop()
                    
                return ret

            elif tag == CompactSerializer.LIST_STREAM_ID:
                
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of list in stream exceeds %d"%limits.INT_MAX)

                ret = []
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    for i in range(n):
                        ret.append(self.readObject(i))
                finally:
                    self.readContext.pop()
                
                return ret

            elif tag == CompactSerializer.SET_STREAM_ID:
                
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of set in stream exceeds %d"%limits.INT_MAX)

                ret = set()
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    for i in range(n):
                        ret.add(self.readObject(i))
                finally:
                    self.readContext.pop()
                
                return ret

            elif tag == CompactSerializer.MAP_STREAM_ID:
                
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of dict in stream exceeds %d"%limits.INT_MAX)

                ret = {}
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    for i in range(n):
                        k = self.readObject(i)
                        v = self.readObject(str(k))
                        ret[k] = v
                finally:
                    self.readContext.pop()
                
                return ret
            
            else:
                raise UnexpectedTagError("Unexpected tag [%02X] reading object field."%tag)


        except Exception as e:
            raise serializer.DeserializationError("Error reading object field [%s]"%self.getReadContext(field)) from e
    
    def readObjectArray(self,field,memberSerialID):
                
        try:
            tag = self.stream.readByte()
        
            while tag == CompactSerializer.OBJECT_FLUSH_STREAM_ID:
                self.readCache.clear()
                tag = self.stream.readByte()

            if tag == CompactSerializer.OBJECT_REF_STREAM_ID:
                
                oid = self.stream.readLong()
                ret = self.readCache[oid]
                return ret
            
            elif tag == CompactSerializer.OBJECT_ARRAY_STREAM_ID:
                
                serialID = self.stream.readUTF()
                serialVersionUID = self.stream.readLong()
                cid = self.stream.readLong()
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of object array in stream exceeds %d"%limits.INT_MAX)

                nci = CompactSerializer.NUMBER_CLASS_INFOS.get(serialID)
                
                if nci is None:

                    cls = self.serialIDs[serialID]

                    if serialID != memberSerialID:
                        raise ValueError("Class with serial ID [%s] is not of required serialID [%s]"%(cls,serialID,memberSerialID))
    
                    if serialVersionUID != cls.getSerialVersionUID() and not serialVersionUID in cls.getOldSerialVersionUIDs():
                        raise ValueError("Class with serial ID [%s] has unknown serialVersioUID [%d]"%(serialID,serialVersionUID))
    
                    self.readClassInfos[cid] = serializer.ClassInfo(serialID,serialVersionUID)
                
                else:
                    cls = nci[0]
                    
                    if serialID != memberSerialID:
                        raise ValueError("Number class with serial ID [%s] is not of required serialID [%s]"%(serialID,memberSerialID))

                    if serialVersionUID != nci[1].serialVersionUID:
                        raise ValueError("Number class with serial ID [%s] has unknown serialVersioUID [%d]"%(serialID,serialVersionUID))
                    
                    self.readClassInfos[cid] = nci[1]
                
                ret = []
                
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    
                    for i in range(n):
                        member = self.readObject(i)
                        if not isinstance(member,cls):
                            raise ValueError("Member of type [%s] is not a subclass of [%s]"%(type(member),cls))
                        ret.append(member)
                
                finally:
                    self.readContext.pop()
                    
                return ret

            elif tag == CompactSerializer.OBJECT_ARRAY_CIREF_STREAM_ID:
                
                cid = self.stream.readLong()
                oid = self.stream.readLong()
                n = self.stream.readInt()
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("Length of object array in stream exceeds %d"%limits.INT_MAX)

                ci = self.readClassInfos[cid]
                
                cls = self.serialIDs[ci.serialID]

                if ci.serialID != memberSerialID:
                    raise ValueError("Class with serial ID [%s] is not of required serial ID [%s]"%(ci.serialID,memberSerialID))
                
                ret = []
                
                self.readCache[oid] = ret
                
                self.readContext.append(field)
                
                try:
                    
                    for i in range(n):
                        member = self.readObject(i)
                        if not isinstance(member,cls):
                            raise ValueError("Member of type [%s] is not a subclass of [%s]"%(type(member),cls))
                        ret.append(member)
                
                finally:
                    self.readContext.pop()
                    
                return ret
            
            else:
                raise UnexpectedTagError("Unexpected tag [%02X] reading object array field."%tag)


        except Exception as e:
            raise serializer.DeserializationError("Error reading object array field [%s]"%self.getReadContext(field)) from e

    
    def writeBoolean(self,field,value):
        try:
            if value is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.BOOLEAN_STREAM_ID)
                self.stream.writeByte(value)
            
        except Exception as e:
            raise serializer.SerializationError("Error writing byte field [%s]"%self.getWriteContext(field)) from e
        
    def writeByte(self,field,value):
        try:
            if value is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.BYTE_STREAM_ID)
                self.stream.writeByte(value)
            
        except Exception as e:
            raise serializer.SerializationError("Error writing byte field [%s]"%self.getWriteContext(field)) from e

    def writeChar(self,field,value):
        try:
            if value is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.CHAR_STREAM_ID)
                self.stream.writeChar(value)
            
        except Exception as e:
            raise serializer.SerializationError("Error writing char field [%s]"%self.getWriteContext(field)) from e

    def writeShort(self,field,i):
        try:
            if i is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.SHORT_STREAM_ID)
                self.stream.writeShort(i)

        except Exception as e:
            raise serializer.SerializationError("Error writing short field [%s]"%self.getWriteContext(field)) from e

    def writeInt(self,field,i):
        try:
            if i is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.INT_STREAM_ID)
                self.stream.writeInt(i)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing int field [%s]"%self.getWriteContext(field)) from e
        
    def writeLong(self,field,i):
        try:
            if i is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.LONG_STREAM_ID)
                self.stream.writeLong(i)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing long field [%s]"%self.getWriteContext(field)) from e

    def writeFloat(self,field,f):
        try:
            if f is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.FLOAT_STREAM_ID)
                self.stream.writeFloat(f)

        except Exception as e:
            raise serializer.SerializationError("Error writing float field [%s]"%self.getWriteContext(field)) from e

    def writeDouble(self,field,d):
        try:
            if d is None:
                self.stream.writeByte(CompactSerializer.NULL_STREAM_ID)
            else:
                self.stream.writeByte(CompactSerializer.DOUBLE_STREAM_ID)
                self.stream.writeDouble(d)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing double field [%s]"%self.getWriteContext(field)) from e

    def writeBooleanArray(self,field,ba):
        try:
            if not self._write_cached(ba):
                self.stream.writeByte(CompactSerializer.BOOLEAN_ARRAY_STREAM_ID)
                self._cache_write(ba)
                self.stream.writeByteArray(ba)
            
        except Exception as e:
            raise serializer.SerializationError("Error writing byte array field [%s]"%self.getWriteContext(field)) from e

    def writeByteArray(self,field,ba):
        try:
            if not self._write_cached(ba):
                self.stream.writeByte(CompactSerializer.BYTE_ARRAY_STREAM_ID)
                self._cache_write(ba)
                self.stream.writeByteArray(ba)
            
        except Exception as e:
            raise serializer.SerializationError("Error writing byte array field [%s]"%self.getWriteContext(field)) from e

    def writeShortArray(self,field,ia):
        try:
            if not self._write_cached(ia):
                self.stream.writeByte(CompactSerializer.SHORT_ARRAY_STREAM_ID)
                self._cache_write(ia)
                self.stream.writeShortArray(ia)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing short array field [%s]"%self.getWriteContext(field)) from e


    def writeIntArray(self,field,ia):
        try:
            if not self._write_cached(ia):
                self.stream.writeByte(CompactSerializer.INT_ARRAY_STREAM_ID)
                self._cache_write(ia)
                self.stream.writeIntArray(ia)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing int array field [%s]"%self.getWriteContext(field)) from e

    def writeLongArray(self,field,ia):
        try:
            if not self._write_cached(ia):
                self.stream.writeByte(CompactSerializer.LONG_ARRAY_STREAM_ID)
                self._cache_write(ia)
                self.stream.writeLongArray(ia)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing long array field [%s]"%self.getWriteContext(field)) from e

    def writeFloatArray(self,field,fa):
        try:
            if not self._write_cached(fa):
                self.stream.writeByte(CompactSerializer.FLOAT_ARRAY_STREAM_ID)
                self._cache_write(fa)
                self.stream.writeFloatArray(fa)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing float array field [%s]"%self.getWriteContext(field)) from e

    def writeDoubleArray(self,field,da):
        try:
            if not self._write_cached(da):
                self.stream.writeByte(CompactSerializer.DOUBLE_ARRAY_STREAM_ID)
                self._cache_write(da)
                self.stream.writeDoubleArray(da)

        except Exception as e:
            raise serializer.SerializationError("Error writing double array field [%s]"%self.getWriteContext(field)) from e

    def writeCharArray(self,field,s):
        try:
            if not self._write_cached(s):
                self.stream.writeByte(CompactSerializer.CHAR_ARRAY_STREAM_ID)
                self._cache_write(s)
                self.stream.writeCharArray(s)
        
        except Exception as e:
            raise serializer.SerializationError("Error writing double array field [%s]"%self.getWriteContext(field)) from e

    def writeObject(self,field,o):
        
        try:
            
            if self._write_cached(o):
                return
            
            if isinstance(o,str):
                if len(o) > 16383:
                    self.stream.writeByte(CompactSerializer.LONG_STRING_STREAM_ID)
                    self._cache_write(o)
                    self.stream.writeUTFLong(o)
                else:
                    self.stream.writeByte(CompactSerializer.STRING_STREAM_ID)
                    self._cache_write(o)
                    self.stream.writeUTF(o)
                return
            
            if isinstance(o,numbers.Number):
                if isinstance(o,float):
                    self.stream.writeByte(CompactSerializer.DOUBLE_STREAM_ID)
                    self.stream.writeDouble(o)
                elif isinstance(o,bool):
                    self.stream.writeByte(CompactSerializer.BOOLEAN_STREAM_ID)
                    self.stream.writeByte(o)
                elif limits.BYTE_MIN <= o and o <= limits.BYTE_MAX:
                    self.stream.writeByte(CompactSerializer.BYTE_STREAM_ID)
                    self.stream.writeByte(o)
                elif limits.SHORT_MIN <= o and o <= limits.SHORT_MAX:
                    self.stream.writeByte(CompactSerializer.SHORT_STREAM_ID)
                    self.stream.writeShort(o)
                elif limits.INT_MIN <= o and o <= limits.INT_MAX:
                    self.stream.writeByte(CompactSerializer.INT_STREAM_ID)
                    self.stream.writeInt(o)
                else:
                    self.stream.writeByte(CompactSerializer.LONG_STREAM_ID)
                    self.stream.writeLong(o)
                return
            
            if isinstance(o,ByteString):
                self.stream.writeByte(CompactSerializer.BYTE_ARRAY_STREAM_ID)
                self._cache_write(o)
                self.stream.writeByteArray(o)
                return
            
            if isinstance(o,serializer.Serializable):
                ci  = serializer.ClassInfo(o.getSerialID(),o.getSerialVersionUID())
                cid = self.writeClassInfos.get(ci)
                
                if cid is None:
                    cid = self._make_cid()
                    self.writeClassInfos[ci] = cid
                    self.stream.writeByte(CompactSerializer.OBJECT_STREAM_ID)
                    self.stream.writeUTF(ci.serialID)
                    self.stream.writeLong(ci.serialVersionUID)
                else:
                    self.stream.writeByte(CompactSerializer.OBJECT_CIREF_STREAM_ID)
                    
                self.stream.writeLong(cid)
                self._cache_write(o)  
                
                self.writeContext.append(field)
                
                try:
                    o.writeFields(self)    
                finally:
                    self.writeContext.pop()
                return
            
            if isinstance(o,Sequence):
                self.stream.writeByte(CompactSerializer.LIST_STREAM_ID)
                self._cache_write(o)
                
                n = len(o)
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("length of list exceeds %d"%limits.INT_MAX)
                
                self.stream.writeInt(n)
                
                self.writeContext.append(field)
                
                try:
                    for i,v in enumerate(o):
                        self.writeObject(i,v)
                        
                finally:
                    self.writeContext.pop()
                return
                
            if isinstance(o,Mapping):
                self.stream.writeByte(CompactSerializer.MAP_STREAM_ID)
                self._cache_write(o)
                
                n = len(o)
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("length of dict exceeds %d"%limits.INT_MAX)
                
                self.stream.writeInt(n)
                
                self.writeContext.append(field)
                
                try:
                    for i,kv in enumerate(o.items()):
                        self.writeObject(i,kv[0])
                        self.writeObject(str(kv[0]),kv[1])
                        
                finally:
                    self.writeContext.pop()
                return
            
            if isinstance(o,Set):
                self.stream.writeByte(CompactSerializer.SET_STREAM_ID)
                self._cache_write(o)
                
                n = len(o)
                
                if n < 0 or n > limits.INT_MAX:
                    raise ValueError("length of set exceeds %d"%limits.INT_MAX)
                
                self.stream.writeInt(n)
                
                self.writeContext.append(field)
                
                try:
                    for i,v in enumerate(o):
                        self.writeObject(i,v)
                        
                finally:
                    self.writeContext.pop()
                return
                
            raise ValueError("Cannot serialize object of type [%s]"%type(o))
            
        except Exception as e:
            raise serializer.SerializationError("Error writing object field [%s]"%self.getWriteContext(field)) from e

    def writeObjectArray(self,field,memberSerialID,o):
        
        if self._write_cached(o):
            return

        n = len(o)
        
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of object array exceeds %d"%limits.INT_MAX)

        nci = CompactSerializer.NUMBER_CLASS_INFOS.get(memberSerialID)
                
        if nci is None:

            cls = self.serialIDs[memberSerialID]
            ci  = serializer.ClassInfo(cls.getSerialID(),cls.getSerialVersionUID())
        else:
            cls = nci[0]
            ci = nci[1]

        cid = self.writeClassInfos.get(ci)
                
        if cid is None:
            cid = self._make_cid()
            self.writeClassInfos[ci] = cid
            self.stream.writeByte(CompactSerializer.OBJECT_ARRAY_STREAM_ID)
            self.stream.writeUTF(ci.serialID)
            self.stream.writeLong(ci.serialVersionUID)
        else:
            self.stream.writeByte(CompactSerializer.OBJECT_ARRAY_CIREF_STREAM_ID)
                    
        self.stream.writeLong(cid)
        self._cache_write(o)
        self.stream.writeInt(n)
                
        self.writeContext.append(field)
                
        try:
            for i,v in enumerate(o):
                self.writeObject(i,v)

        finally:
            self.writeContext.pop()

    def flush(self):
        self.stream.writeByte(CompactSerializer.OBJECT_FLUSH_STREAM_ID)
        self.writeCache.clear()
        self.stream.flush()
    
    def close(self):
        self.stream.close()