'''
..
  $Id$
  
  Remoting python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 04.02.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

from clazzes.remoting import cmd
from clazzes.remoting import compact
from clazzes.remoting import aux
from clazzes.remoting import throwable
from clazzes.remoting import uid
from clazzes.tools.lang import extension

import types
import socketserver
import logging

log = logging.getLogger(__name__)

class ConnectionHandler(socketserver.BaseRequestHandler):
    '''
      Handle a client connection on the server side and delegate incoming
      InvocationCmd instances to a set of given singleton bean instances.
    '''
    
    def __init__(self):
        '''
          Set up a connection handler based on a serializer and a mapping of
          bsubsystem names to exported singleton beans.
        '''
        self.uid = None
        self.srlzr = None
    
    def _check_interrupt(self):
        with self.server.interrupt_lock:
            return self.server.interrupted
    
    def handle(self):
        
        ca = self.client_address[0] if len(self.client_address) > 0 else '<local>'
        
        self.uid = uid.UID.newInstance()
        
        log.info("Got connection from [%s] with uid [%s]"%(ca,self.uid))

        self.srlzr = compact.CompactSerializer(self.request,self.server.serialIDs)

        try:
            
            uidcmd = cmd.ExchangeClientUIDCmd()
            uidcmd.clientUID = self.uid
            self.srlzr.writeObject("__exchangeClientUID",uidcmd)

            while not self._check_interrupt():
            
                cmd = self.srlzr.readObject("__cmd")
                
                cmd.receive_server(self)

        except Exception as e:
            log.error("Connection from [%s] with uid [%s] terminated by exception: %s"%(ca,self.uid,e))
            
        finally:
            self.srlzr.close()
    
    def _invoke(self,subSystem:str,arg:aux.RemoteInvocation):
        
        bean = self.server.beans.get(subSystem)
        
        if bean is None:
            raise aux.RemoteInvocationFailureException("Unknown subsystem [%s]"%subSystem)
        
        if not isinstance(arg,aux.RemoteInvocation):
            raise aux.RemoteInvocationFailureException("Received invocation nor of type [%s]"%aux.RemoteInvocation.getSerialID())
            
        method = bean.get(arg.methodName)
        
        if method is None or not isinstance(method,types.MethodType):
            raise aux.RemoteInvocationFailureException("Subsystem [%s] has no method named [%s]."%(subSystem,arg.methodName))
        
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("Invoking method [%s] on subsystem [%s] with arguments [%s]...",(arg.methodName,subSystem,arg.methodName))
        
        try:
            ret = method(arg.arguments)
        
            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("Method [%s] on subsystem [%s] returned [%s].",(arg.methodName,subSystem,ret))

            return ret

        except Exception as e:
            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("Method [%s] on subsystem [%s] raised exception of type [%s].",(arg.methodName,subSystem,type(e)))

            raise throwable.FlatRemoteException(type(e),str(e))
        
@extension(cmd.InvocationCmd)
def receive_server(self,handler:ConnectionHandler):
    
    try:
        ret = handler._invoke(self.subSystem,self.arg)
        
        irc = cmd.InvocationResultCmd()
        irc.result = ret
        
        handler.srlzr.writeObject("__invocationResult",irc)
        
    except throwable.Throwable as e:
        
        iec = cmd.InvocationExceptionCmd()
        iec.exception = e
        
        handler.srlzr.writeObject("__invocationException",iec)
        