'''
..
  $Id$

  Remoting python tools of the clazzes.org project
  http://www.clazzes.org

  Created: 04.02.2018

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import abc
import logging
import types
from collections.abc import Mapping

log = logging.getLogger(__name__)


def build_serialIDs(*args):
    '''
      Build a dictionary of serialIDs to classes from a list of dictionaries or
      python packages
    '''
    serialIDs = {}

    for packageOrDict in args:
        if isinstance(packageOrDict,types.ModuleType):
            if hasattr(packageOrDict,"SERIAL_IDS"):
                serialIDs.update(packageOrDict.SERIAL_IDS)
            else:
                log.warn("Ignoring module [%s] without SERIAL_IDS property.",packageOrDict.__name__)

        elif isinstance(packageOrDict,Mapping):
            serialIDs.update(packageOrDict)

        else:
            log.warn("Ignoring parameter of unsupported type [%s] (mappings or modules expected).",type(packageOrDict))

class SerializationError(Exception):
    pass

class DeserializationError(Exception):
    pass

class Serializer(metaclass=abc.ABCMeta):
    '''
    The interface class for a serializer implementation.
    '''

    @abc.abstractmethod
    def readBoolean(self,field):
        '''
        Read a 1 byte boolean.
        :param field: The name of the field to read.
        :return: A boolean instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readByte(self,field):
        '''
        Read an 8 bit integer.
        :param field: The name of the field to read.
        :return: An integer in the range from -128 to 127.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readShort(self,field):
        '''
        Read a 16 bit integer.
        :param field: The name of the field to read.
        :return: An integer in the range from -32768 to 32767.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readInt(self,field):
        '''
        Read a 32 bit integer.
        :param field: The name of the field to read.
        :return: An integer in the range from -2147483648 to 2147483647.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readLong(self,field):
        '''
        Read a 64 bit integer.
        :param field: The name of the field to read.
        :return: An integer in the range from -9223372036854775808 to 9223372036854775807.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readChar(self,field):
        '''
        Read a 16 bit character.
        :param field: The name of the field to read.
        :return: A string of length 1 with a unicode code point between 0 and 65535.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readFloat(self,field):
        '''
        Read a 32 bit single precision IEEE 754 floating point number.
        :param field: The name of the field to read.
        :return: A float instance.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readDouble(self,field):
        '''
        Read a 64 bit double precision IEEE 754 floating point number.
        :param field: The name of the field to read.
        :return: A float instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readBooleanArray(self,field):
        '''
        Read a boolean array.
        :param field: The name of the field to read.
        :return: An array of boolean instances.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readByteArray(self,field):
        '''
        Read an byte array.
        :param field: The name of the field to read.
        :return: A bytes instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readShortArray(self,field):
        '''
        Read a fixed-size array of 16 bit integers.
        :param field: The name of the field to read.
        :return: A list of integers in the range from -32768 to 32767.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readIntArray(self,field):
        '''
        Read an array of 32 bit integers.
        :param field: The name of the field to read.
        :return: A list of integers in the range from -2147483648 to 2147483647.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readLongArray(self,field):
        '''
        Read an array 64 bit integer.
        :param field: The name of the field to read.
        :return: A list of integers in the range from -9223372036854775808 to 9223372036854775807.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readCharArray(self,field):
        '''
        Read an array of 16 bit characters aka a string.
        :param field: The name of the field to read.
        :return: A string of composed of unicode code points between 0 and 65535.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readFloatArray(self,field):
        '''
        Read an array of 32 bit single precision IEEE 754 floating point numbers.
        :param field: The name of the field to read.
        :return: A list of float instances.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readDoubleArray(self,field):
        '''
        Read an array of 64 bit double precision IEEE 754 floating point number.
        :param field: The name of the field to read.
        :return: A list of float instances.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readObject(self,field):
        '''
        Read an object from the stream based on the next tag found in the stream.
        :param field: The name of the field to read.
        :return: An object read from the stream.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readObjectArray(self,field,memberSerialID):
        '''
        Read an object array from the stream.
        :param field: The name of the field to read.
        :param memberSerialID: The serial ID of the class of the members, which might be java.lang.Object for plain object arrays.
        :return: An object read from the stream.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeBoolean(self,field,b):
        '''
        Write an 1 byte booelan.
        :param field: The name of the field to read.
        :param b: A boolean instance to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeByte(self,field,b):
        '''
        Write an 8 bit integer.
        :param field: The name of the field to read.
        :param b: An integer in the range from -128 to 127 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeShort(self,field,i):
        '''
        Write a 16 bit integer.
        :param field: The name of the field to read.
        :param i: An integer in the range from -32768 to 32767 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeInt(self,field,i):
        '''
        Write a 32 bit integer.
        :param field: The name of the field to read.
        :param i: An integer in the range from -2147483648 to 2147483647 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeLong(self,field,i):
        '''
        Write a 64 bit integer.
        :param field: The name of the field to read.
        :param i: An integer in the range from -9223372036854775808 to 9223372036854775807 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeChar(self,field,c):
        '''
        Write the first character of the given string.
        :param field: The name of the field to read.
        :param c: A string to write the first character of.
                  The unicode code point must be in the range from 0 to 65535.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeFloat(self,field,f):
        '''
        Write a 32 bit single-precision IEEE 754 floating point number.
        :param field: The name of the field to read.
        :param f: A floating point number.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeDouble(self,field,d):
        '''
        Write a 64 bit double-precision IEEE 754 floating point number.
        :param field: The name of the field to read.
        :param d: A floating point number.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeBooleanArray(self,field,ba):
        '''
        Write an array of booleans.
        :param field: The name of the field to read.
        :param ba: A list instance with boolean members.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeByteArray(self,field,ba):
        '''
        Write an array of 8 bit integers or a bytes-like object.
        :param field: The name of the field to read.
        :param ba: A bytes instance of a list of integers in the range from -128 to 127.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeShortArray(self,field,ia):
        '''
        Write an array of 16 bit integer.
        :param field: The name of the field to read.
        :param ia: A list of integers in the range from -32768 to 32767 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeIntArray(self,field,ia):
        '''
        Write an array of 32 bit integer.
        :param field: The name of the field to read.
        :param i: A list of integers in the range from -2147483648 to 2147483647 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeLongArray(self,field,ia):
        '''
        Write an array of 64 bit integer.
        :param field: The name of the field to read.
        :param ia: A list of integers in the range from -9223372036854775808 to 9223372036854775807 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeCharArray(self,field,s):
        '''
        Write a character array with code points in the range from 0 to 65535.
        :param field: The name of the field to read.
        :param s: A string to write the characters of.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeFloatArray(self,field,fa):
        '''
        Write an array of 32 bit single-precision IEEE 754 floating point numbers.
        :param field: The name of the field to read.
        :param fa: A list of floating point numbers.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeDoubleArray(self,field,da):
        '''
        Write an arry  64 bit double-precision IEEE 754 floating point numbers.
        :param field: The name of the field to read.
        :param da: A list of floating point numbers.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeObject(self,field,o):
        '''
        Write an object to the stream based on the type of the passed object.
        List instances are written as a list to the stream. If list instances represent
        numeric arrays, use on of the methods writeXyzArray().
        :param field: The name of the field to write.
        :param o: The object to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeObjectArray(self,field,memberSerialID,o):
        '''
        Write an object array to the stream.
        :param field: The name of the field to read.
        :param memberSerialID: The serial ID of the class of the members, which might be java.lang.Object for plain object arrays.
        :param o: A list of objects to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def flush(self):
        '''
        Flush all write caches and the output stream.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def close(self):
        '''
        Close this serializer and the underlying transport.
        '''
        raise NotImplemented()

    def __enter__(self):
        return self

    def __exit__(self,*args):  # @UnusedVariable
        self.close()
        return False

class ClassInfo:
    def __init__(self,serialID : str, serialVersionUID: int):
        self.serialID = serialID
        self.serialVersionUID = serialVersionUID

    def __eq__(self,other):
        return self.serialID == other.serialID and self.serialVersionUID == other.serialVersionUID

    def __ne__(self,other):
        return self.serialID != other.serialID or self.serialVersionUID != other.serialVersionUID

    def __lt__(self,other):
        if self.serialID < other.serialID:
            return True
        elif self.serialID > other.serialID:
            return False
        else:
            return self.serialVersionUID < other.serialVersionUID

    def __le__(self,other):
        if self.serialID < other.serialID:
            return True
        elif self.serialID > other.serialID:
            return False
        else:
            return self.serialVersionUID <= other.serialVersionUID

    def __gt__(self,other):
        if self.serialID > other.serialID:
            return True
        elif self.serialID < other.serialID:
            return False
        else:
            return self.serialVersionUID > other.serialVersionUID

    def __ge__(self,other):
        if self.serialID > other.serialID:
            return True
        elif self.serialID < other.serialID:
            return False
        else:
            return self.serialVersionUID >= other.serialVersionUID

    def __hash__(self):
        return hash(self.serialID) ^ self.serialVersionUID

class Serializable(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def getSerialID():
        '''
        :return: The serial ID aka the JAVA class name of this class.
        '''
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def getSerialVersionUID():
        '''
        :return: The 64bit serial version UID of this class version.
        '''
        raise NotImplementedError()

    @staticmethod
    def getOldSerialVersionUIDs():
        '''
        :return: The supported legacy 64bit serial version UIDs of this class version.
        '''
        return tuple()

    @abc.abstractmethod
    def readFields(self,serializer:Serializer,serialVersionUID:int):
        '''
        :param serializer: The serializer to read fields from.
        :param serialVersionUID: The serial version UID of the stream version, which
                  might be a legacy serial version UID with an outdated set of fields.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeFields(self,serializer:Serializer):
        '''
        :param serializer: The serializer to write fields to.
        '''
        raise NotImplementedError()

class JavaObject(Serializable):
    '''
    Pseudo serializable class to be used as component for object arrays.
    '''
    @staticmethod
    def getSerialID():
        return "java.lang.Object"

    @staticmethod
    def getSerialVersionUID():
        return -1

class JavaEnumMeta(abc.ABCMeta):

    def __init__(self,name,bases,clsdict):  # @UnusedVariable
        for i,v in enumerate(clsdict["_VALUES"]):
            setattr(self,v,self(i,v))

class JavaEnum(JavaObject,metaclass = JavaEnumMeta):

    _VALUES = ()

    def __new__(cls,*args):

        if len(args) == 0:
            return getattr(cls,cls._VALUES[0])
        elif len(args) == 1:
            v = args[0]
            if isinstance(v,int):
                return getattr(cls,cls._VALUES[v])
            elif isinstance(v,str):
                return getattr(cls,v)
            else:
                raise ValueError("Cannot instantiate JavaEnum instance from type [%s]"%type(v))
        elif len(args) == 2:

            if getattr(cls,args[1],None) != None:
                raise ValueError("Cannot instantiate JavaEnum instance with multiple argument after metaclass initialization.")

            ret = JavaObject.__new__(cls)
            ret.value = args[0]
            ret.name = args[1]
            return ret

        elif len(args) == 2:
            raise ValueError("Cannot instantiate JavaEnum instance with with more then 2 arguments.")

    '''
    Base class for serializable JAVA enum objects.
    '''
    @staticmethod
    def getSerialID():
        return "java.lang.Enum"

    @staticmethod
    def getSerialVersionUID():
        return -1

    def __str__(self):
        return self.name

    @classmethod
    def readInstance(cls:type,serializer:Serializer,serialVersionUID:int):  # @UnusedVariable
        return cls(serializer.readInt("v"))

    def readFields(self,serializer:Serializer,serialVersionUID:int):  # @UnusedVariable
        raise NotImplementedError("Use static JavaEnum.readInstance()")

    def writeFields(self,serializer:Serializer):
        serializer.writeInt("v",self.value)

    def __gt__(self,other):
        return self.value > other.value

    def __ge__(self,other):
        return self.value >= other.value

    def __lt__(self,other):
        return self.value < other.value

    def __le__(self,other):
        return self.value <= other.value

SERIAL_IDS = {
    JavaObject.getSerialID(): JavaObject,
    JavaEnum.getSerialID(): JavaEnum
}