'''
..
  $Id$
  
  Remoting python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 04.02.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

from clazzes.remoting import limits

import abc
import xdrlib
import logging

from collections.abc import ByteString

log = logging.getLogger(__name__)

class DataStream(metaclass=abc.ABCMeta):
    '''
    The interface class for binary datastream implementations.
    '''
    
    @abc.abstractmethod
    def readByte(self):
        '''
        Read an 8 bit integer.
        :return: An integer in the range from -128 to 127.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readShort(self):
        '''
        Read a 16 bit integer.
        :return: An integer in the range from -32768 to 32767.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readInt(self):
        '''
        Read a 32 bit integer.
        :return: An integer in the range from -2147483648 to 2147483647.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readLong(self):
        '''
        Read a 64 bit integer.
        :return: An integer in the range from -9223372036854775808 to 9223372036854775807.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readChar(self):
        '''
        Read a 16 bit character.
        :return: A string of length 1 with a unicode code point between 0 and 65535. 
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readFloat(self):
        '''
        Read a 32 bit sigle precision IEEE 754 floating point number.
        :return: A float instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readDouble(self):
        '''
        Read a 64 bit double precision IEEE 754 floating point number.
        :return: A float instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readUTF(self):
        '''
        Read a UTF-8 encoded string with up to 65535 bytes.
        :return: A string.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readUTFLong(self):
        '''
        Read a UTF-8 encoded string with up to 2147483647 bytes.
        :return: A string.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readByteArray(self):
        '''
        Read a byte array prefixed by an integer length..
        :return: A bytes instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readShortArray(self):
        '''
        Read a fixed-size array of 16 bit integers.
        :return: A list of integers in the range from -32768 to 32767.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readIntArray(self):
        '''
        Read an array of 32 bit integers.
        :return: A list of integers in the range from -2147483648 to 2147483647.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readLongArray(self):
        '''
        Read an array 64 bit integer.
        :return: A list of integers in the range from -9223372036854775808 to 9223372036854775807.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readCharArray(self):
        '''
        Read an array of 16 bit characters aka a string.
        :return: A string of composed of unicode code points between 0 and 65535. 
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def readFloatArray(self):
        '''
        Read an array of 32 bit single precision IEEE 754 floating point numbers.
        :return: A list of float instance.
        '''
        raise NotImplementedError()


    @abc.abstractmethod
    def readDoubleArray(self):
        '''
        Read an array of 64 bit double precision IEEE 754 floating point number.
        :return: A list of float instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeByte(self,b):
        '''
        Write an 8 bit integer.
        :param b: An integer in the range from -128 to 127 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeShort(self,i):
        '''
        Write a 16 bit integer.
        :param i: An integer in the range from -32768 to 32767 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeInt(self,i):
        '''
        Write a 32 bit integer.
        :param i: An integer in the range from -2147483648 to 2147483647 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeLong(self,i):
        '''
        Write a 64 bit integer.
        :param i: An integer in the range from -9223372036854775808 to 9223372036854775807 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeChar(self,c):
        '''
        Write the first character of the given string.
        :param c: A string to write the first character of.
                  The unicode code point must be in the range from 0 to 65535.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeFloat(self,f):
        '''
        Write a 32 bit single-precision IEEE 754 floating point number.
        :param f: A floating point number.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeDouble(self,i):
        '''
        Write a 64 bit double-precision IEEE 754 floating point number.
        :param i: A floating point number.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeUTF(self,s):
        '''
        Write a UTF-8 encoded string with up to 65535 bytes.
        :param s: A string to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeUTFLong(self,s):
        '''
        Write a UTF-8 encoded string with up to 2147483647 bytes.
        :param s: A string to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeByteArray(self,ba):
        '''
        Write a bytes-like object prefixed by an integer length.
        :param ba: A bytes-like object instance.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeShortArray(self,ia):
        '''
        Write an array of 16 bit integer.
        :param ia: A list of integers in the range from -32768 to 32767 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeIntArray(self,ia):
        '''
        Write an array of 32 bit integer.
        :param i: A list of integers in the range from -2147483648 to 2147483647 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeLongArray(self,ia):
        '''
        Write an array of 64 bit integer.
        :param ia: A list of integers in the range from -9223372036854775808 to 9223372036854775807 to write.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeCharArray(self,s):
        '''
        Write a character array with code points in the range from 0 to 65535.
        :param s: A string to write the characters of.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeFloatArray(self,fa):
        '''
        Write an array of 32 bit single-precision IEEE 754 floating point numbers.
        :param fa: A list of floating point numbers.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def writeDoubleArray(self,da):
        '''
        Write an arry  64 bit double-precision IEEE 754 floating point numbers.
        :param da: A list of floating point numbers.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def flush(self):
        '''
        Flush the output stream.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def close(self):
        '''
        Close the stream and the underlying transport
        '''
        raise NotImplementedError()

class BinaryDataStream(DataStream):
    '''
    A binary data stream operating on a binary file handle an reading/writing
    numbers in network byte order aka MSB-first.
    '''
    
    def __init__(self,fh):
        '''
        :param fh: The binary file-like object to read from and/or write to.
        '''
        self.fh = fh
        self.packer = None
    
    def _get_packer(self):
        
        if self.packer is None:
            self.packer = xdrlib.Packer()
        else:
            self.packer.reset()
        
        return self.packer
    
    def readByte(self):

        b = self.fh.read(1)
        
        if len(b) != 1:
            raise EOFError('Unexpected end of data stream')
        
        i = b[0]
        
        if i & 0x80:
            return i - 0x100
        else:
            return i

    def readChar(self):
        b = self.fh.read(2)
        
        if len(b) != 2:
            raise EOFError('Unexpected end of data stream')
        
        return chr(b[0] << 8 | b[1])

    def readShort(self):
        b = self.fh.read(2)
        
        if len(b) != 2:
            raise EOFError('Unexpected end of data stream')
        
        i = b[0] << 8 | b[1]
        
        if i & 0x8000:
            return i - 0x10000
        else:
            return i

    def readInt(self):
        b = self.fh.read(4)
        
        if len(b) != 4:
            raise EOFError('Unexpected end of data stream')
        
        i = (b[0]&0x7f) << 24 | b[1] << 16 | b[2] << 8 | b[3]
    
        if b[0] & 0x80:
            return i - 0x80000000
        else:
            return i
        
    def readLong(self):
        b = self.fh.read(8)
        
        if len(b) != 8:
            raise EOFError('Unexpected end of data stream')
        
        i = (b[0]&0x7f) << 56 | b[1] << 48 | b[2] << 40 | b[3] << 32 | b[4] << 24 | b[5] << 16 | b[6] << 8 | b[7]
    
        if b[0] & 0x80:
            return i - 0x8000000000000000
        else:
            return i
        
    def readFloat(self):
        b = self.fh.read(4)
        
        if len(b) != 4:
            raise EOFError('Unexpected end of data stream')
        
        return xdrlib.Unpacker(b).unpack_float()
        
    def readDouble(self):
        b = self.fh.read(8)
            
        if len(b) != 8:
            raise EOFError('Unexpected end of data stream')
        
        return xdrlib.Unpacker(b).unpack_double()

    def readUTF(self):
        bn = self.fh.read(2)
        
        if len(bn) != 2:
            raise EOFError('Unexpected end of data stream')
        
        n = bn[0] << 8 | bn[1]
        
        b = self.fh.read(n)
        
        if len(b) != n:
            raise EOFError('Unexpected end of data stream')

        return b.decode('utf-8')
        
    def readUTFLong(self):
        
        l = self.readInt()
        if l < 0 or l > limits.INT_MAX:
            raise ValueError("UTF-16 length of long string exceeds %d"%limits.INT_MAX)

        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("UTF-8 length of long string exceeds %d"%limits.INT_MAX)
        
        b = self.fh.read(n)
        
        if len(b) != n:
            raise EOFError('Unexpected end of data stream')
        
        s = b.decode('utf-8')
        if len(s) != l:
            log.warn("Warning unicode string length [%d] does not match announced string length [%d]"%(len(s),l))
        return s

    def readByteArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of byte array exceeds %d"%limits.INT_MAX)

        ba = self.fh.read(n)
        
        if len(ba) != n:
            raise EOFError('Unexpected end of data stream')
    
        return ba
    
    def readShortArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of short array exceeds %d"%limits.INT_MAX)

        ret = []
        
        for i in range(n):  # @UnusedVariable
            ret.append(self.readShort())
        
        return ret

    def readIntArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of int array exceeds %d"%limits.INT_MAX)

        ret = []
        
        for i in range(n):  # @UnusedVariable
            ret.append(self.readInt())
        
        return ret

    def readLongArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of long array exceeds %d"%limits.INT_MAX)

        ret = []
        
        for i in range(n):  # @UnusedVariable
            ret.append(self.readLong())
        
        return ret

    def readFloatArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of float array exceeds %d"%limits.INT_MAX)

        ret = []
        
        for i in range(n):  # @UnusedVariable
            ret.append(self.readFloat())
        
        return ret

    def readDoubleArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of double array exceeds %d"%limits.INT_MAX)

        ret = []
        
        for i in range(n):  # @UnusedVariable
            ret.append(self.readDouble())
        
        return ret

    def readCharArray(self):
        n = self.readInt()
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("length of char array exceeds %d"%limits.INT_MAX)

        ret = ""
        
        for i in range(n):  # @UnusedVariable
            ret += self.readChar()
        
        return ret

    def writeByte(self,value):
        self.fh.write(bytes((value&0xff,)))

    def writeChar(self,value):
        i = ord(value)
        self.fh.write(bytes((i>>8 & 0xff, i & 0xff)))
        
    def writeShort(self,i):
        self.fh.write(bytes((i>>8 & 0xff, i & 0xff)))

    def writeInt(self,i):
        self.fh.write(bytes((i>>24 & 0xff, i>>16 & 0xff, i>>8 & 0xff, i & 0xff)))
        
    def writeLong(self,i):
        self.fh.write(bytes((i>>56 & 0xff,i>>48 & 0xff,i>>40 & 0xff,i>>32 & 0xff,
                             i>>24 & 0xff,i>>16 & 0xff,i>>8 & 0xff,i & 0xff)))

    def writeFloat(self,f):
        packer = self._get_packer()
        packer.pack_float(f)
        self.fh.write(packer.get_buffer())

    def writeDouble(self,d):
        packer = self._get_packer()
        packer.pack_double(d)
        self.fh.write(packer.get_buffer())

    def writeUTF(self,s):
        b = s.encode('utf-8')
        n = len(b)
        if n > 65535:
            raise ValueError("UTF-8 length of string exceeds 65535")
        
        self.writeShort(n)
        self.fh.write(b)
        
    def writeUTFLong(self,s):
        b = s.encode('utf-8')
        l = len(s)
        n = len(b)
        if n > limits.INT_MAX:
            raise ValueError("UTF-8 length of string exceeds %d"%limits.INT_MAX)
        
        self.fh.write(bytes((l>>24 & 0xff, l>>16 & 0xff, l>>8 & 0xff, l & 0xff,
                             n>>24 & 0xff, n>>16 & 0xff, n>>8 & 0xff, n & 0xff)))
        self.fh.write(b)
    
    def writeByteArray(self,ba):
        
        n = len(ba)
        if n > limits.INT_MAX:
            raise ValueError("length of byte array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
            
        if isinstance(ba,ByteString):
            self.fh.write(ba)
        else:
            i=0
            while i+4096 < n:
                self.fh.write(bytes([ba[i] & 0xff for i in range(i,i+4096)]))
                i+=4096
            
            if i < n:
                self.fh.write(bytes([ba[i] & 0xff for i in range(i,n)]))
    
    def writeShortArray(self,ia):
        n = len(ia)
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("Length of short array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
        for i in ia:
            self.writeShort(i)

    def writeIntArray(self,ia):

        n = len(ia)
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("Length of int array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
        for i in ia:
            self.writeInt(i)

    def writeLongArray(self,ia):
      
        n = len(ia)
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("Length of long array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
        for i in ia:
            self.writeLong(i)

    def writeFloatArray(self,fa):
        n = len(fa)
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("Length of float array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
        for f in fa:
            self.writeFloat(f)

    def writeDoubleArray(self,da):
        n = len(da)
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("Length of double array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
        for d in da:
            self.writeDouble(d)

    def writeCharArray(self,s):
        n = len(s)
        if n < 0 or n > limits.INT_MAX:
            raise ValueError("Length of char array exceeds %d"%limits.INT_MAX)

        self.writeInt(n)
        for c in s:
            self.writeChar(c)

    def flush(self):
        self.fh.flush()

    def close(self):
        self.fh.close()
        
    def __enter__(self):
        return self

    def __exit__(self,*args): # @UnusedVariable
        self.close()
        return False