'''
..
  $Id$

  Remoting python tools of the clazzes.org project
  http://www.clazzes.org

  Created: 04.02.2018

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import io
from clazzes.remoting import serializer

class StackTraceElement(serializer.JavaObject):
    __slots__ = ["declaringClass","fileName","lineNumber","methodName"]

    def __init__(self):
        self.declaringClass = None
        self.fileName = None
        self.lineNumber = 0
        self.methodName = None

    def __str__(self):
        if self.fileName is None:
            return "%s.%s"%(self.declaringClass,self.methodName)
        else:
            return "%s.%s (%s:%d)"%(self.declaringClass,self.methodName,self.fileName,self.lineNumber)

    @staticmethod
    def getSerialID():
        return "java.lang.StackTraceElement"

    @staticmethod
    def getSerialVersionUID():
        return 6992337162326171013

    def readFields(self, srlzr:serializer.Serializer, serialVersionUID:int):  # @UnusedVariable
        self.declaringClass = srlzr.readObject("declaringClass")
        self.fileName = srlzr.readObject("fileName")
        self.lineNumber = srlzr.readInt("lineNumber")
        self.methodName = srlzr.readObject("methodName")

    def writeFields(self, srlzr:serializer.Serializer):
        srlzr.writeObject("declaringClass",self.declaringClass)
        srlzr.writeObject("fileName",self.fileName)
        srlzr.writeInt("lineNumber",self.lineNumber)
        srlzr.writeObject("methodName",self.methodName)

class Throwable(serializer.JavaObject,Exception):

    __slots__ = [ "cause","detailMessage","stackTrace" ]

    def __init__(self,detailMessage=None,cause=None,stackTrace=None):
        self.cause = cause
        self.detailMessage = detailMessage
        self.stackTrace = stackTrace

    def writeToTextIO(self,os):

        os.write("%s: %s"%(self.getSerialID(),self.detailMessage))

        if self.stackTrace:
            for st in self.stackTrace:
                os.write("\n")
                os.write("  ")
                os.write(str(st))

        if self.cause and self.cause != self:
            os.write("\n\ncaused by:\n")
            self.cause.writeToTextIO(os)

    def __str__(self):

        os = io.StringIO()
        self.writeToTextIO(os)
        return os.getvalue()

    @staticmethod
    def getSerialID():
        return "java.lang.Throwable"

    @staticmethod
    def getSerialVersionUID():
        return -3042686055658047285

    def readFields(self, srlzr:serializer.Serializer, serialVersionUID:int):  # @UnusedVariable
        self.cause = srlzr.readObject("cause")
        self.detailMessage = srlzr.readObject("detailMessage")
        self.stackTrace = srlzr.readObjectArray("stackTrace",StackTraceElement.getSerialID())

    def writeFields(self, srlzr:serializer.Serializer):
        pass

class JavaException(Throwable):

    @staticmethod
    def getSerialID():
        return "java.lang.Exception"

    @staticmethod
    def getSerialVersionUID():
        return -3387516993124229948


class RuntimeException(JavaException):

    @staticmethod
    def getSerialID():
        return "java.lang.RuntimeException"

    @staticmethod
    def getSerialVersionUID():
        return -7034897190745766939

class FlatRemoteException(JavaException):

    __slots__ = JavaException.__slots__ + ["originalClassName"]

    def __init__(self,originalClassName=None,detailMessage=None,cause=None,stackTrace=None):
        super().__init__(detailMessage,cause,stackTrace)
        self.originalClassName = originalClassName

    @staticmethod
    def getSerialID():
        return "org.clazzes.remoting.FlatRemoteException"

    @staticmethod
    def getSerialVersionUID():
        return -7831517399321247574

    def readFields(self, srlzr:serializer.Serializer, serialVersionUID:int):  # @UnusedVariable
        self.originalClassName = srlzr.readObject("originalClassName")
        JavaException.readFields(self,srlzr,JavaException.getSerialVersionUID())

    def writeFields(self, srlzr:serializer.Serializer):
        srlzr.writeObject("originalClassName",self.originalClassName)
        JavaException.writeFields(self, srlzr)

SERIAL_IDS = {
    Throwable.getSerialID(): Throwable,
    JavaException.getSerialID(): JavaException,
    RuntimeException.getSerialID(): RuntimeException,
    FlatRemoteException.getSerialID(): FlatRemoteException,
    StackTraceElement.getSerialID(): StackTraceElement
}