'''
..
  $Id$
  
  Remoting python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 04.02.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

from clazzes.remoting import serializer
from clazzes.remoting import limits

import socket
import threading
import time
import uuid

class UID(serializer.JavaObject):
    '''
    :ivar unique: A 32bit unique host identifier, which is equal for all UID instance
                  created during a single python run.
    :ivar time: A 64 bit timestamp in milliseconds-since-the-ecpoch,
                which is is only issued once per python run.
    :ivar count: A 16 bit count 
    '''
    
    _lock = threading.Lock()
    _hostunique = None
    _next_count = 0
    _last_time = 0
    
    def __init__(self,unique=limits.INT_MIN,time=limits.LONG_MIN,count=limits.SHORT_MIN):
        '''
        Initialize an UID instance by it's fields.
        For generating a unique instance, call UID.newInstance()
        '''
        self.count = count
        self.time = time
        self.unique = unique
    
    @staticmethod
    def newInstance():
        '''
        Return an instance, which is initialized in a way, that it's fields are unique.
        Uniqueness is defined by the hostname, process Id and the time of creation of the
        UID measured in millseconds since the epoch 1970-01-01 00:00:00 GMT.
        '''
        
        with UID._lock:
            if UID._hostunique is None:
                h = hash(socket.getfqdn())
                if h >= 0:
                    UID._hostunique = h & limits.INT_MAX
                else:
                    UID._hostunique = limits.INT_MIN + (h & limits.INT_MAX)
                    
            next_time = int(time.time() * 1000)
            if next_time <= UID._last_time:
                UID._last_time += 1
            else:
                UID._last_time = next_time;
            
            UID._next_count += 1
            if UID._next_count > limits.SHORT_MAX:
                UID._next_count = limits.SHORT_MIN
            
            ret = UID(UID._hostunique,UID._last_time,UID._next_count);
            return ret
    
    @staticmethod
    def getSerialID():
        return "java.rmi.server.UID"
    
    @staticmethod
    def getSerialVersionUID():
        return 1086053664494604050
    
    def readFields(self, srlzr:serializer.Serializer, serialVersionUID:int):  # @UnusedVariable
        self.count  = srlzr.readShort("count")
        self.time   = srlzr.readLong("time")
        self.unique = srlzr.readInt("unique")

    def writeFields(self, srlzr:serializer.Serializer):
        srlzr.writeShort("count",self.count)
        srlzr.writeLong("time",self.time)
        srlzr.writeInt("unique",self.unique)

    def __str__(self):
        return "%08x:%016x:%04x"%(self.unique,self.time,self.count)
    
    def __hash__(self):
        return self.unique ^ self.time ^ self.count
    
    def __eq__(self,other):
        return self.unique == other.unique and self.time == other.time and self.count == other.count
    
    def __neq__(self,other):
        return self.unique != other.unique or self.time != other.time or self.count != other.count

class JavaUUID(serializer.JavaObject,uuid.UUID):
    
    @staticmethod
    def newInstance():
        u = uuid.uuid4()
        return JavaUUID(int=u.int)
    
    @staticmethod
    def getSerialID():
        return "java.util.UUID"
    
    @staticmethod
    def getSerialVersionUID():
        return -4856846361193249489
    
    @classmethod
    def readInstance(cls:type,srlzr:serializer.Serializer,serialVersionUID:int):  # @UnusedVariable

        leastSigBits = srlzr.readLong("leastSigBits")
        mostSigBits = srlzr.readLong("mostSigBits")
        
        b = bytes((mostSigBits>>56 & 0xff,mostSigBits>>48 & 0xff,mostSigBits>>40 & 0xff,mostSigBits>>32 & 0xff,
                   mostSigBits>>24 & 0xff,mostSigBits>>16 & 0xff,mostSigBits>>8 & 0xff,mostSigBits & 0xff,
                   leastSigBits>>56 & 0xff,leastSigBits>>48 & 0xff,leastSigBits>>40 & 0xff,leastSigBits>>32 & 0xff,
                   leastSigBits>>24 & 0xff,leastSigBits>>16 & 0xff,leastSigBits>>8 & 0xff,leastSigBits & 0xff))
        
        return JavaUUID(bytes=b)

    def readFields(self,srlzr:serializer.Serializer,serialVersionUID:int):  # @UnusedVariable
        raise NotImplementedError("Use static JavaUUID.readInstance()")

    def writeFields(self, srlzr:serializer.Serializer):
        
        b=self.bytes
        
        i = (b[0]&0x7f) << 56 | b[1] << 48 | b[2] << 40 | b[3] << 32 | b[4] << 24 | b[5] << 16 | b[6] << 8 | b[7]
    
        if b[0] & 0x80:
            mostSigBits = i - 0x8000000000000000
        else:
            mostSigBits = i

        i = (b[8]&0x7f) << 56 | b[9] << 48 | b[10] << 40 | b[11] << 32 | b[12] << 24 | b[13] << 16 | b[14] << 8 | b[15]
    
        if b[8] & 0x80:
            leastSigBits = i - 0x8000000000000000
        else:
            leastSigBits = i
          
        srlzr.writeLong("leastSigBits",leastSigBits)
        srlzr.writeLong("mostSigBits",mostSigBits)


SERIAL_IDS = {
    UID.getSerialID(): UID,
    JavaUUID.getSerialID(): JavaUUID
}